"""Tag model."""

from django.core.cache import cache
from django.db import IntegrityError, models, transaction
from django.db.models import Max
from django.utils import timezone
from django.utils.text import slugify

from djpress.conf import settings as djpress_settings

TAG_CACHE_KEY = "tags"


class TagManager(models.Manager):
    """Tag manager."""

    def get_queryset(self) -> models.QuerySet:
        """Return the default queryset for tag."""
        return super().get_queryset()

    def get_tags(self) -> models.QuerySet:
        """Return the queryset for tags.

        This should be used instead of the default queryset to get the tags since it will check the cache first if
        needed.

        If CACHE_TAGS is set to True, we return the cached queryset. Or if there's no cached queryset, we cache it
        and then return it.
        """
        if djpress_settings.CACHE_TAGS:
            return self._get_cached_tags()

        return self.get_queryset()

    def _get_cached_tags(self) -> models.QuerySet:
        """Return the cached tags queryset."""
        queryset = cache.get(TAG_CACHE_KEY)

        if queryset is None:
            queryset = self.get_queryset()
            cache.set(TAG_CACHE_KEY, queryset, timeout=None)

        return queryset

    def get_tag_by_slug(self, slug: str) -> "Tag":
        """Return a single tag by its slug.

        If tag caching is enabled, we first try to get the tag from the cache. If the tag is not found in the cache,
        we fetch it from the database.

        Args:
            slug (str): The slug of the tag.

        Returns:
            Tag: The tag with the given slug.

        Raises:
            ValueError: If the tag is not found.
        """
        tags = self.get_tags()
        tag = next(
            (tag for tag in tags if tag.slug == slug),
            None,
        )

        # If no tag is found, either it doesn't exist, or if caching is enabled, it's not in the cache.
        if not tag:
            # If caching is not enabled, we raise a ValueError.
            if not djpress_settings.CACHE_TAGS:
                msg = "Tag not found"
                raise ValueError(msg)

            # If caching is enabled, we try fetch the tag from the database.
            try:
                tag = self.get(slug=slug)
            except Tag.DoesNotExist as exc:
                msg = "Tag not found"
                # Raise a 404 error
                raise ValueError(msg) from exc

        return tag

    def get_tags_with_published_posts(self) -> models.QuerySet:
        """Return a queryset of tags that have published posts.

        We can use the has_posts property to include only tags with published posts.
        """
        return Tag.objects.filter(pk__in=[tag.pk for tag in self.get_queryset() if tag.has_posts])


class Tag(models.Model):
    """Tag model."""

    title = models.CharField(max_length=100, unique=True)
    slug = models.SlugField(max_length=100, unique=True)

    # Custom manager
    objects: "TagManager" = TagManager()

    class Meta:
        """Meta options for the tag model."""

        verbose_name = "Tag"
        verbose_name_plural = "Tags"

    def __str__(self) -> str:
        """Return the title of the tag."""
        return self.title

    def save(self: "Tag", *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        """Override the save method to auto-generate the slug."""
        if not self.slug:
            self.slug = slugify(self.title)
            if not self.slug or self.slug.strip("-") == "":
                msg = "Invalid title. Unable to generate a valid slug."
                raise ValueError(msg)

        try:
            with transaction.atomic():
                super().save(*args, **kwargs)
        except IntegrityError as exc:
            msg = f"A tag with the slug {self.slug} already exists."
            raise ValueError(msg) from exc

    @property
    def url(self) -> str:
        """Return the tag's URL."""
        from djpress.url_utils import get_tag_url

        return get_tag_url(self)

    @property
    def posts(self) -> models.QuerySet:
        """Return only published posts.

        TODO: duplicated logic for what a published post is. Need a better way of managing this.
        """
        return self._posts.filter(
            post_type="post",
            status="published",
            date__lte=timezone.now(),
        )

    @property
    def has_posts(self) -> bool:
        """Return True if the tag has published posts."""
        return self.posts.exists()

    @property
    def last_modified(self) -> None | timezone.datetime:
        """Return the most recent last modified date of posts in the tag.

        This property is used in the sitemap to determine the last modified date of the tag.

        If the tag has no published posts, we return None.

        Returns:
            None | timezone.datetime: The most recent last modified date of posts in the tag.
        """
        return self.posts.aggregate(latest=Max("modified_date"))["latest"]



from django.contrib.contenttypes.models import ContentType
import pytest

from django.contrib.auth.models import Group
from django.contrib.auth.models import Permission

from djpress.models import Post


@pytest.fixture
def post_content_type() -> ContentType:
    return ContentType.objects.get_for_model(Post)


@pytest.mark.django_db
def test_groups_created() -> None:
    # This should run post_migrate signal
    from django.core.management import call_command

    call_command("migrate")

    # Check groups exist
    assert Group.objects.filter(name="editor").exists()
    assert Group.objects.filter(name="author").exists()
    assert Group.objects.filter(name="contributor").exists()


@pytest.mark.django_db
def test_group_permissions(post_content_type: ContentType) -> None:
    editor = Group.objects.get(name="editor")
    author = Group.objects.get(name="author")
    contributor = Group.objects.get(name="contributor")

    # Check publish permission
    publish_perm = Permission.objects.get(content_type=post_content_type, codename="can_publish_post")
    assert publish_perm in editor.permissions.all()
    assert publish_perm in author.permissions.all()
    assert publish_perm not in contributor.permissions.all()


@pytest.mark.django_db
def test_group_category_permissions(post_content_type: ContentType) -> None:
    editor = Group.objects.get(name="editor")
    author = Group.objects.get(name="author")
    contributor = Group.objects.get(name="contributor")

    # Check category permissions
    category_perms = Permission.objects.filter(
        content_type=post_content_type, codename__in=["add_category", "change_category", "delete_category"]
    )
    for perm in category_perms:
        assert perm in editor.permissions.all()
        assert perm not in author.permissions.all()
        assert perm not in contributor.permissions.all()



"""Signals for djpress app."""

from django.apps import AppConfig
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache
from django.db.models.signals import post_delete, post_migrate, post_save
from django.dispatch import receiver

from djpress.models.category import (
    CATEGORY_CACHE_KEY,
    Category,
)
from djpress.models.post import (
    PUBLISHED_POSTS_CACHE_KEY,
    Post,
)
from djpress.models.tag import (
    TAG_CACHE_KEY,
    Tag,
)


@receiver(post_migrate)
def create_groups(sender: AppConfig, **_) -> None:  # noqa: ANN003
    """Create groups and assign permissions."""
    if sender.name != "djpress":
        return

    # Get content types
    post_content_type = ContentType.objects.get_for_model(Post)
    category_content_type = ContentType.objects.get_for_model(Category)

    # Get permissions
    standard_permissions = Permission.objects.filter(
        content_type=post_content_type,
        codename__in=["add_post", "change_post", "delete_post"],
    )
    publish_permission = Permission.objects.get(
        content_type=post_content_type,
        codename="can_publish_post",
    )
    category_permissions = Permission.objects.filter(
        content_type=category_content_type,
        codename__in=["add_category", "change_category", "delete_category"],
    )

    # Create groups and assign permissions
    editor_group, _ = Group.objects.get_or_create(name="editor")
    editor_group.permissions.add(publish_permission, *standard_permissions, *category_permissions)

    author_group, _ = Group.objects.get_or_create(name="author")
    author_group.permissions.add(publish_permission, *standard_permissions)

    contributor_group, _ = Group.objects.get_or_create(name="contributor")
    contributor_group.permissions.add(*standard_permissions)


@receiver(post_save, sender=Category)
@receiver(post_delete, sender=Category)
def invalidate_category_cache(**_) -> None:  # noqa: ANN003
    """Invalidate the category cache.

    We invalidate the cache when a category is saved or deleted.
    """
    cache.delete(CATEGORY_CACHE_KEY)


@receiver(post_save, sender=Post)
@receiver(post_delete, sender=Post)
def invalidate_published_content_cache(**_) -> None:  # noqa: ANN003
    """Invalidate the published posts cache.

    We invalidate the cache when a post is saved or deleted.
    """
    cache.delete(PUBLISHED_POSTS_CACHE_KEY)


@receiver(post_save, sender=Tag)
@receiver(post_delete, sender=Tag)
def invalidate_tag_cache(**_) -> None:  # noqa: ANN003
    """Invalidate the tag cache.

    We invalidate the cache when a tag is saved or deleted.
    """
    cache.delete(TAG_CACHE_KEY)



"""djpress admin configuration."""

from django.contrib import admin
from django.db.models.query import QuerySet
from django.http import HttpRequest
from django.utils import timezone
from django.utils.html import format_html

# Register the models here.
from djpress.models import Category, PluginStorage, Post, Tag


@admin.register(Tag)
class TagAdmin(admin.ModelAdmin):
    """Tag admin configuration."""

    list_display = ["title", "slug"]
    prepopulated_fields = {"slug": ("title",)}
    search_fields = ["title", "slug"]


@admin.register(Category)
class CategoryAdmin(admin.ModelAdmin):
    """Category admin configuration."""

    list_display = ["title", "slug"]


@admin.register(Post)
class PostAdmin(admin.ModelAdmin):
    """Post admin configuration."""

    list_display = ["post_type", "published_status", "title", "parent", "formatted_date", "author"]
    list_display_links = ["title"]
    ordering = ["post_type", "-date"]  # Displays pages first, then sorted by date.
    list_filter = ["post_type", "date", "author"]
    prepopulated_fields = {"slug": ("title",)}
    search_fields = ["title", "content", "slug"]
    readonly_fields = ["modified_date"]
    fieldsets = [
        (
            None,
            {
                "fields": ("title", "slug", "author"),
            },
        ),
        (
            "Content",
            {
                "fields": ("content",),
                "classes": ("wide",),
            },
        ),
        (
            "Publishing",
            {
                "fields": (
                    "status",
                    "post_type",
                    "date",
                    "modified_date",
                ),
            },
        ),
        (
            "Post Settings",
            {
                "fields": ("categories", "tags"),
                "description": "These settings only apply to posts",
            },
        ),
        (
            "Page Settings",
            {
                "fields": ("parent", "menu_order"),
                "description": "These settings only apply to pages",
            },
        ),
    ]

    @admin.display(
        boolean=True,
        description="Published",
    )
    def published_status(self, obj: Post) -> bool:
        """Display published status using Django's built-in boolean icon."""
        return obj.is_published

    def formatted_date(self, obj: Post) -> str:
        """Format the date and show future posts distinctly."""
        if obj.date > timezone.now():
            return format_html(
                '<span style="color: #666;">{} (Scheduled)</span>',
                obj.date.strftime("%Y-%m-%d %H:%M"),
            )
        return obj.date.strftime("%Y-%m-%d %H:%M")

    formatted_date.short_description = "Date"

    def get_queryset(self, request: HttpRequest) -> QuerySet[Post]:
        """Limit the queryset based on user role.

        - Superusers and Editors can see all posts.
        - Contributors and Authors can only see their own posts.

        Args:
            request: The request object.

        Returns:
            A queryset of posts.
        """
        qs = super().get_queryset(request)

        # Superusers can see all posts
        if request.user.is_superuser:
            return qs

        # Editors can see all posts
        if request.user.groups.filter(name="editor").exists():
            return qs

        # Authors and Contributors see only their own posts
        return qs.filter(author=request.user)

    def has_change_permission(self, request: HttpRequest, obj: Post = None) -> bool:
        """Limit the change permission based on user role.

        Args:
            request: The request object.
            obj: The post object.

        Returns:
            A boolean indicating if the user has change permission.
        """
        # First check if they have basic change permission
        if not super().has_change_permission(request, obj):
            return False

        # Superusers can change any post
        if request.user.is_superuser:
            return True

        # If no specific object, they have general change permission
        if obj is None:
            return True

        # Editors can change any post
        if request.user.groups.filter(name="editor").exists():
            return True

        # Others can only change their own posts
        return obj.author == request.user

    def get_readonly_fields(self, request: HttpRequest, _: Post = None) -> tuple[str, ...]:
        """Limit the readonly fields based on user role.

        Args:
            request: The request object.
            _: The post object.

        Returns:
            A tuple of readonly fields.
        """
        # Superusers can edit all fields
        if request.user.is_superuser:
            return self.readonly_fields

        # Restrict status field if user can't publish
        if not request.user.has_perm("djpress.can_publish_post"):
            return (*self.readonly_fields, "status")

        return self.readonly_fields


@admin.register(PluginStorage)
class PluginStorageAdmin(admin.ModelAdmin):
    """PluginStorage admin configuration."""

    list_display = ["plugin_name", "plugin_data"]
    list_display_links = ["plugin_name"]
    ordering = ["plugin_name"]
    search_fields = ["plugin_name"]
    list_filter = ["plugin_name"]



"""Models package for djpress app."""

from djpress.models.category import Category
from djpress.models.plugin_storage import PluginStorage
from djpress.models.post import Post
from djpress.models.tag import Tag

__all__ = ["Category", "PluginStorage", "Post", "Tag"]



import pytest

from django import test
from django.conf.locale import te
from django.core.cache import cache
from django.utils.text import slugify
from django.utils import timezone

from djpress.models.tag import Tag, TAG_CACHE_KEY


@pytest.mark.django_db
def test_tag_model(tag1):
    tag = Tag.objects.create(title="Test Tag", slug="test-tag")
    assert tag.title == "Test Tag"
    assert tag.slug == "test-tag"
    assert str(tag) == "Test Tag"


@pytest.mark.django_db
def test_tag_save_slug_generation():
    """Test that the slug is correctly generated when saving a Tag."""
    tag = Tag(title="Test Tag")
    tag.save()

    assert tag.slug == slugify("Test Tag")


@pytest.mark.django_db
def test_tag_save_slug_uniqueness():
    """Test that an error is raised when trying to save a Tag with a duplicate slug."""
    tag1 = Tag(title="Test Tag")
    tag1.save()

    tag2 = Tag(title="Test Tag")

    with pytest.raises(ValueError) as excinfo:
        tag2.save()

    assert str(excinfo.value) == f"A tag with the slug {tag2.slug} already exists."


@pytest.mark.django_db
def test_tag_save_invalid_name():
    """Test that an error is raised when trying to save a Tag with an invalid name."""
    tag = Tag(title="-")

    with pytest.raises(ValueError) as excinfo:
        tag.save()

    assert str(excinfo.value) == "Invalid title. Unable to generate a valid slug."


@pytest.mark.django_db
def test_tag_slug_auto_generation():
    # Test case 1: Slug auto-generated when not provided
    tag1 = Tag.objects.create(title="Test Tag")
    assert tag1.slug == slugify(tag1.title)

    # Test case 2: Slug not overridden when provided
    tag2 = Tag.objects.create(title="Another Tag", slug="custom-slug")
    assert tag2.slug == "custom-slug"

    # Test case 3: Slug auto-generated with special characters
    tag3 = Tag.objects.create(title="Special !@#$%^&*() Tag")
    assert tag3.slug == "special-tag"

    # Test case 4: Slug auto-generated with non-ASCII characters
    tag4 = Tag.objects.create(title="Non-ASCII áéíóú Tag")
    assert tag4.slug == "non-ascii-aeiou-tag"

    # Test case 5: Slug auto-generated with leading/trailing hyphens
    tag5 = Tag.objects.create(title="--Leading/Trailing Hyphens--")
    assert tag5.slug == "leadingtrailing-hyphens"

    # Test case 6: Raise ValueError for invalid title
    with pytest.raises(ValueError) as exc_info:
        Tag.objects.create(title="!@#$%^&*()")
    assert str(exc_info.value) == "Invalid title. Unable to generate a valid slug."


@pytest.mark.django_db
def test_get_tags_cache_enabled(settings, tag1, tag2, tag3):
    """Test that the get_tags method returns the correct tags."""
    # Confirm the settings in settings_testing.py
    settings.DJPRESS_SETTINGS["CACHE_TAGS"] = True
    assert settings.DJPRESS_SETTINGS["CACHE_TAGS"] is True

    tags = Tag.objects.get_tags()

    assert list(tags) == [tag1, tag2, tag3]


@pytest.mark.django_db
def test_get_tags_cache_disabled(settings, tag1, tag2, tag3):
    """Test that the get_tags method returns the correct tags."""
    # Confirm the settings in settings_testing.py
    settings.DJPRESS_SETTINGS["CACHE_TAGS"] = False
    assert settings.DJPRESS_SETTINGS["CACHE_TAGS"] is False

    tags = Tag.objects.get_tags()

    assert list(tags) == [tag1, tag2, tag3]


@pytest.mark.django_db
def test_get_cached_tags(settings, tag1, tag2, tag3):
    """Test that the _get_cached_tags method returns the correct tags."""
    # Confirm the settings in settings_testing.py
    settings.DJPRESS_SETTINGS["CACHE_TAGS"] = False
    assert settings.DJPRESS_SETTINGS["CACHE_TAGS"] is False

    tags = Tag.objects.get_tags()
    cache.set(TAG_CACHE_KEY, tags)

    settings.DJPRESS_SETTINGS["CACHE_TAGS"] = True
    assert settings.DJPRESS_SETTINGS["CACHE_TAGS"] is True

    tags = Tag.objects._get_cached_tags()
    assert list(tags) == [tag1, tag2, tag3]

    # Now the queryset should be cached
    cached_queryset = cache.get(TAG_CACHE_KEY)
    assert list(tags) == list(cached_queryset)


@pytest.mark.django_db
def test_get_tag_by_slug_cache_enabled(settings, tag1, tag2):
    """Test that the get_tag_by_slug method returns the correct tag."""
    # Confirm the settings in settings_testing.py
    settings.DJPRESS_SETTINGS["CACHE_TAGS"] = True
    assert settings.DJPRESS_SETTINGS["CACHE_TAGS"] is True

    tag = Tag.objects.get_tag_by_slug(tag1.slug)

    assert tag == tag1
    assert not tag == tag2

    with pytest.raises(ValueError) as excinfo:
        tag = Tag.objects.get_tag_by_slug("non-existent-tag")


@pytest.mark.django_db
def test_get_category_by_slug_cache_disabled(settings, tag1, tag2):
    """Test that the get_tag_by_slug method returns the correct category."""
    # Confirm the settings in settings_testing.py
    settings.DJPRESS_SETTINGS["CACHE_TAGS"] = False
    assert settings.DJPRESS_SETTINGS["CACHE_TAGS"] is False

    tag = Tag.objects.get_tag_by_slug(tag1.slug)

    assert tag == tag1
    assert not tag == tag2

    with pytest.raises(ValueError) as excinfo:
        tag = Tag.objects.get_tag_by_slug("non-existent-tag")


@pytest.mark.django_db
def test_get_tag_by_slug_not_exists(settings):
    """Test that the get_tag_by_slug method returns None when the tag does not exist."""
    # Confirm the settings in settings_testing.py
    settings.DJPRESS_SETTINGS["CACHE_TAGS"] = False
    assert settings.DJPRESS_SETTINGS["CACHE_TAGS"] is False

    with pytest.raises(ValueError) as excinfo:
        _ = Tag.objects.get_tag_by_slug("non-existent-tag")
    assert "Tag not found" in str(excinfo.value)


@pytest.mark.django_db
def test_category_url(settings):
    """Test that the url property returns the correct URL."""
    # Confirm the settings in settings_testing.py
    assert settings.DJPRESS_SETTINGS["TAG_ENABLED"] is True
    assert settings.DJPRESS_SETTINGS["TAG_PREFIX"] == "test-url-tag"

    tag = Tag.objects.create(title="Test Tag", slug="test-tag")

    assert tag.url == "/test-url-tag/test-tag/"

    settings.DJPRESS_SETTINGS["TAG_ENABLED"] = False
    settings.DJPRESS_SETTINGS["TAG_PREFIX"] = ""

    assert tag.url == "/test-tag/"


@pytest.mark.django_db
def test_tag_posts(test_post1, test_post2, tag1, tag2):
    test_post1.tags.add(tag1)
    test_post2.tags.add(tag2)
    assert list(tag1.posts.all()) == [test_post1]
    assert list(tag2.posts.all()) == [test_post2]

    test_post2.tags.add(tag1)
    assert list(tag1.posts.all()) == [test_post1, test_post2]
    assert list(tag2.posts.all()) == [test_post2]


@pytest.mark.django_db
def test_tag_has_posts(test_post1, test_post2, tag1, tag2):
    assert tag1.has_posts is False
    assert tag2.has_posts is False

    test_post1.tags.add(tag1)
    assert tag1.has_posts is True
    assert tag2.has_posts is False

    test_post2.tags.add(tag2)
    assert tag1.has_posts is True
    assert tag2.has_posts is True

    test_post1.status = "draft"
    test_post1.save()
    assert tag1.has_posts is False
    assert tag2.has_posts is True

    test_post2.date = timezone.now() + timezone.timedelta(days=1)
    test_post2.save()
    assert tag1.has_posts is False
    assert tag2.has_posts is False


@pytest.mark.django_db
def test_get_tag_published(test_post1, test_post2, tag1, tag2):
    assert list(Tag.objects.get_tags_with_published_posts()) == []

    test_post1.tags.add(tag1)
    test_post2.tags.add(tag2)
    assert list(Tag.objects.get_tags_with_published_posts()) == [tag1, tag2]

    test_post1.status = "draft"
    test_post1.save()
    assert list(Tag.objects.get_tags_with_published_posts()) == [tag2]

    test_post2.date = timezone.now() + timezone.timedelta(days=1)
    test_post2.save()
    assert list(Tag.objects.get_tags_with_published_posts()) == []


@pytest.mark.django_db
def test_tag_last_modified(test_post1, test_post2, tag1, tag2):
    test_post1.tags.add(tag1)
    test_post2.tags.add(tag2)

    assert tag1.last_modified == test_post1.modified_date
    assert tag2.last_modified == test_post2.modified_date

    test_post1.modified_date = timezone.now() + timezone.timedelta(days=1)
    test_post1.save()
    assert tag1.last_modified == test_post1.modified_date

    test_post1.status = "draft"
    test_post1.save()
    assert tag1.last_modified is None



"""Tests for the PostAdmin class."""

from django import test
from django.contrib.auth.models import Group
from django.test import RequestFactory
import pytest
from django.contrib.admin import site
from django.utils import timezone
from datetime import timedelta

from djpress.admin import PostAdmin
from djpress.models import Post
from djpress.url_utils import User
from tests.conftest import superuser


@pytest.fixture
def request_factory() -> RequestFactory:
    return RequestFactory()


@pytest.fixture
def post_admin():
    return PostAdmin(Post, site)


@pytest.mark.django_db
def test_published_status(test_post1):
    """Test the published_status method returns correct boolean values."""
    # Get the admin class
    post_admin = PostAdmin(Post, site)

    # Test published post
    assert post_admin.published_status(test_post1) is True

    # Chnage to draft
    test_post1.status = "draft"
    test_post1.save()
    assert post_admin.published_status(test_post1) is False

    # Change to future post
    test_post1.status = "published"
    test_post1.date = timezone.now() + timedelta(days=1)
    test_post1.save()
    assert post_admin.published_status(test_post1) is False


@pytest.mark.django_db
def test_formatted_date(test_post1):
    """Test the formatted_date method returns the correct date."""
    # Get the admin class
    post_admin = PostAdmin(Post, site)

    # Test formatted date
    assert post_admin.formatted_date(test_post1) == test_post1.date.strftime("%Y-%m-%d %H:%M")

    # Test future post
    test_post1.date = timezone.now() + timedelta(days=1)
    test_post1.save()

    assert (
        post_admin.formatted_date(test_post1)
        == f'<span style="color: #666;">{test_post1.date.strftime("%Y-%m-%d %H:%M")} (Scheduled)</span>'
    )


@pytest.mark.django_db
def test_queryset_superuser(
    test_post1: Post,
    test_post2: Post,
    test_post3: Post,
    request_factory: RequestFactory,
    superuser: User,
    post_admin: PostAdmin,
) -> None:
    """Test the get_queryset method for superusers."""

    # Create a request object
    request = request_factory.get("/")
    request.user = superuser

    # Test get_queryset method
    assert post_admin.get_queryset(request=request).count() == 3


@pytest.mark.django_db
def test_queryset_editor(
    test_post1: Post,
    test_post2: Post,
    test_post3: Post,
    user: User,
    post_admin: PostAdmin,
    request_factory: RequestFactory,
) -> None:
    """Test the get_queryset method for editors."""
    # Create an editor user
    editor_group = Group.objects.get(name="editor")
    user.groups.clear()  # Remove the user from all other groups
    user.groups.add(editor_group)

    # Create a request object
    request = request_factory.get("/")
    request.user = user

    qs = post_admin.get_queryset(request).count() == 3


@pytest.mark.django_db
def test_queryset_author(
    test_post1: Post,
    test_post2: Post,
    test_post3: Post,
    user: User,
    post_admin: PostAdmin,
    request_factory: RequestFactory,
) -> None:
    """Test the get_queryset method for authors.

    Authors should only see their own posts.
    """
    # Create an editor user
    author_group = Group.objects.get(name="author")
    user.groups.clear()  # Remove the user from all other groups
    user.groups.add(author_group)

    # Create a request object
    request = request_factory.get("/")
    request.user = user

    qs = post_admin.get_queryset(request).count() == 0

    test_post1.author = user
    test_post1.save()

    qs = post_admin.get_queryset(request).count() == 1

    test_post2.author = user
    test_post2.save()

    qs = post_admin.get_queryset(request).count() == 2


@pytest.mark.django_db
def test_queryset_contributor(
    test_post1: Post,
    test_post2: Post,
    test_post3: Post,
    user: User,
    post_admin: PostAdmin,
    request_factory: RequestFactory,
) -> None:
    """Test the get_queryset method for contributors.

    Contributors should only see their own posts.
    """
    # Create an editor user
    contributor_group = Group.objects.get(name="contributor")
    user.groups.clear()  # Remove the user from all other groups
    user.groups.add(contributor_group)

    # Create a request object
    request = request_factory.get("/")
    request.user = user

    qs = post_admin.get_queryset(request).count() == 0

    test_post1.author = user
    test_post1.save()

    qs = post_admin.get_queryset(request).count() == 1

    test_post2.author = user
    test_post2.save()

    qs = post_admin.get_queryset(request).count() == 2


@pytest.mark.django_db
def test_has_change_permission_superuser(
    superuser: User,
    post_admin: PostAdmin,
    request_factory: RequestFactory,
    test_post1: Post,
) -> None:
    request = request_factory.get("/")
    request.user = superuser

    assert post_admin.has_change_permission(request) is True  # General permission
    assert post_admin.has_change_permission(request, test_post1) is True  # Can edit any post


@pytest.mark.django_db
def test_has_change_permission_editor(
    user: User,
    post_admin: PostAdmin,
    request_factory: RequestFactory,
    test_post1: Post,
) -> None:
    """Test the has_change_permission method for editors.

    Editors can edit any post.
    """
    # Create an editor user that makes the request
    editor_user = User.objects.create_user(username="editor_user")
    editor_group = Group.objects.get(name="editor")
    editor_user.groups.clear()  # Remove the user from all other groups
    editor_user.groups.add(editor_group)
    request = request_factory.get("/")
    request.user = editor_user

    assert post_admin.has_change_permission(request) is True  # General permission
    assert post_admin.has_change_permission(request, test_post1) is True  # Can edit any post

    test_post1.author = editor_user
    test_post1.save()

    assert post_admin.has_change_permission(request, test_post1) is True  # Can edit own post


@pytest.mark.django_db
def test_has_change_permission_author(
    post_admin: PostAdmin,
    request_factory: RequestFactory,
    test_post1: Post,
) -> None:
    """Test the has_change_permission method for authors.

    Authors can only edit their own posts.
    """
    # Create an author user that makes the request
    author_user = User.objects.create_user(username="author_user")
    author_group = Group.objects.get(name="author")
    author_user.groups.clear()  # Remove the user from all other groups
    author_user.groups.add(author_group)
    request = request_factory.get("/")
    request.user = author_user

    assert post_admin.has_change_permission(request) is True  # General permission
    assert post_admin.has_change_permission(request, test_post1) is False  # Cannot edit any post

    test_post1.author = author_user
    test_post1.save()

    assert post_admin.has_change_permission(request, test_post1) is True  # Can edit own post


@pytest.mark.django_db
def test_has_change_permission_contributor(
    post_admin: PostAdmin,
    request_factory: RequestFactory,
    test_post1: Post,
) -> None:
    """Test the has_change_permission method for contributors.

    Authors can only edit their own posts.
    """
    # Create an contributor user that makes the request
    contributor_user = User.objects.create_user(username="contributor_user")
    contributor_group = Group.objects.get(name="contributor")
    contributor_user.groups.clear()  # Remove the user from all other groups
    contributor_user.groups.add(contributor_group)
    request = request_factory.get("/")
    request.user = contributor_user

    assert post_admin.has_change_permission(request) is True  # General permission
    assert post_admin.has_change_permission(request, test_post1) is False  # Cannot edit any post

    test_post1.author = contributor_user
    test_post1.save()

    assert post_admin.has_change_permission(request, test_post1) is True  # Can edit own post


@pytest.mark.django_db
def test_has_change_permission_regular_user(
    post_admin: PostAdmin,
    request_factory: RequestFactory,
    test_post1: Post,
) -> None:
    """Test the has_change_permission method for a regular user.

    A regular user that doesn't belong to any groups won't be able to edit.
    """
    # Create an regular user that makes the request
    regular_user = User.objects.create_user(username="regular_user")
    regular_user.groups.clear()  # Remove the user from all other groups
    request = request_factory.get("/")
    request.user = regular_user

    assert post_admin.has_change_permission(request) is False  # General permission
    assert post_admin.has_change_permission(request, test_post1) is False  # Cannot edit any post

    # Even if they are an author of a post, they can't edit it without permissions.
    test_post1.author = regular_user
    test_post1.save()

    assert post_admin.has_change_permission(request, test_post1) is False  # Can edit own post


@pytest.mark.django_db
def test_get_readonly_fields_superuser(superuser: User, post_admin: PostAdmin, request_factory: RequestFactory):
    request = request_factory.get("/")
    request.user = superuser

    readonly_fields = post_admin.get_readonly_fields(request)
    assert "status" not in readonly_fields  # Superuser can edit status


@pytest.mark.django_db
def test_get_readonly_fields_editor(user: User, post_admin: PostAdmin, request_factory: RequestFactory):
    editor_group = Group.objects.get(name="editor")
    user.groups.add(editor_group)
    request = request_factory.get("/")
    request.user = user

    readonly_fields = post_admin.get_readonly_fields(request)
    assert "status" not in readonly_fields  # Editor can edit status


@pytest.mark.django_db
def test_get_readonly_fields_author(user: User, post_admin: PostAdmin, request_factory: RequestFactory):
    author_group = Group.objects.get(name="author")
    user.groups.add(author_group)
    request = request_factory.get("/")
    request.user = user

    readonly_fields = post_admin.get_readonly_fields(request)
    assert "status" not in readonly_fields  # Author can edit status


@pytest.mark.django_db
def test_get_readonly_fields_contributor(user: User, post_admin: PostAdmin, request_factory: RequestFactory):
    contributor_group = Group.objects.get(name="contributor")
    user.groups.add(contributor_group)
    request = request_factory.get("/")
    request.user = user

    readonly_fields = post_admin.get_readonly_fields(request)
    assert "status" in readonly_fields  # Contributor cannot edit status



# Generated by Django 5.1.4 on 2025-01-23 10:23

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("djpress", "0009_alter_post_options"),
    ]

    operations = [
        migrations.CreateModel(
            name="Tag",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("title", models.CharField(max_length=100, unique=True)),
                ("slug", models.SlugField(max_length=100, unique=True)),
            ],
            options={
                "verbose_name": "Tag",
                "verbose_name_plural": "Tags",
            },
        ),
        migrations.AddField(
            model_name="post",
            name="tags",
            field=models.ManyToManyField(blank=True, related_name="_posts", to="djpress.tag"),
        ),
    ]



# Generated by Django 5.1.3 on 2024-12-16 04:08

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("djpress", "0008_alter_post_options_alter_post_managers_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="post",
            options={
                "default_manager_name": "admin_objects",
                "permissions": [("can_publish_post", "Can publish post")],
                "verbose_name": "post",
                "verbose_name_plural": "posts",
            },
        ),
    ]
